import unittest
from unittest.mock import patch
import dtflw.com as com
import tests.utils as utils
import dtflw.arguments as a
from ddt import ddt, data, unpack


@ddt
class MessageBusTestCase(unittest.TestCase):

    @patch("dtflw.databricks.get_spark_session")
    def test_share_topic_exists(self, get_session_mock):

        # Arrange
        get_session_mock.return_value = utils.SparkSessionMock()
        bus = com.MessageBus()
        bus.share("channel_1", "topic_1", "foo")

        # Act
        bus.share("channel_1", "topic_1", "bar")

        # Assert
        actual = bus.try_get()
        self.assertDictEqual(
            {
                "channel_1": {
                    "topic_1": "bar"
                }
            },
            actual
        )

    @patch("dtflw.databricks.get_spark_session")
    def test_share_channel_exists(self, get_session_mock):

        # Arrange
        get_session_mock.return_value = utils.SparkSessionMock()
        bus = com.MessageBus()
        bus.share("channel_1", "topic_1", "foo")

        # Act
        bus.share("channel_1", "topic_2", "bar")

        # Assert
        actual = bus.try_get()
        self.assertDictEqual(
            {
                "channel_1": {
                    "topic_1": "foo",
                    "topic_2": "bar"
                }
            },
            actual
        )

    @data(
        (None, None, None, {
            "channel_1": {
                "topic_11": "foo",
                "topic_12": "bar"
            },
            'channel_2': {
                "topic_21": "baz"
            }
        }),
        ("channel_1", None, None, {
            "topic_11": "foo",
            "topic_12": "bar"
        }),
        ("channel_2", "topic_21", None, "baz"),
        # Returns default
        ("channel_1", "topic_unknown", 42, 42),
        ("channel_unknown", "topic_1", 42, 42),
    )
    @unpack
    @patch("dtflw.databricks.get_spark_session")
    def test_try_get(self, channel, topic, default, expected, get_session_mock):

        # Arrange
        get_session_mock.return_value = utils.SparkSessionMock()
        bus = com.MessageBus()
        bus.share("channel_1", "topic_11", "foo")
        bus.share("channel_1", "topic_12", "bar")
        bus.share("channel_2", "topic_21", "baz")

        # Act
        actual = bus.try_get(channel, topic, default)

        # Assert
        self.assertEqual(expected, actual)



import types
from typing import Union
from dtflw.flow_context import FlowContext
from dtflw.lazy_notebook import LazyNotebook
from dtflw.plugin import FlowPluginBase, NotebookPluginBase
from functools import partial

class Flow:
    """
    Represents the main object for building a data pipeline.
    """

    def __init__(self, ctx: FlowContext):
        if ctx is None:
            raise ValueError("Context cannot be None.")

        self.__ctx = ctx
        self.__nb_plugins = {}

    def __setitem__(self, _):
        raise NotImplementedError("Not supported.")

    def __getitem__(self, table_name: str):
        """
        Returns the latest evaluation of a table's file path.
        If not found then returns None.

        Parameters
        ----------
        table_name: str
            Name of a table.
        """
        return self.ctx.tables_repo.resolve(table_name)

    @property
    def ctx(self) -> FlowContext:
        """
        Returns the current context.
        """
        return self.__ctx

    def notebook(self, path: str) -> LazyNotebook:
        """
        Adds a notebook to a pipeline.

        Parameters
        ----------
        path: str
            Path to a notebook.
        """
        new_nb = LazyNotebook(path, self.__ctx)

        # Attach plugin actions to a notebook.
        for plugin in self.__nb_plugins.values():
            
            p = partial(plugin.act, new_nb, self)
            setattr(
                new_nb,
                plugin.action_name,
                types.MethodType(
                    partial(plugin.act, new_nb),
                    new_nb
                ))

        return new_nb

    def show(self):
        """
        Prints the current state.
        """
        self.__ctx.show()

    def install(self, plugin: Union[FlowPluginBase, NotebookPluginBase]) -> None:
        """
        Installs a plugin to the current instance of Flow.

        Parameters
        ----------
        plugin : FlowPluginBase | NotebookPluginBase
            A plugin object to be installed.
        """
        if isinstance(plugin, FlowPluginBase):
            setattr(
                self,
                plugin.action_name,
                types.MethodType(plugin.act, self))

        elif isinstance(plugin, NotebookPluginBase):
            if plugin.action_name in self.__nb_plugins:
                raise ValueError(
                    f"Notebook plugin {plugin.action_name} is already installed.")
            self.__nb_plugins[plugin.action_name] = plugin

        else:
            raise ValueError("Unexpected plugin type.")



class NotebookRun():
    """
    Carries information related to a run of a notebook.

    Parameters
    ----------
    notebook_path: str
        Absolute path to a notebook in a worspace or repo.

    args: {str: str}
        A dictionary of arguments.

    inputs: {str, str}
        A dictionary of input tables' names to absolute path they are stored.

    outputs: {str, str}
        A dictionary of outputs tables' names to absolute path they are stored.
    """

    def __init__(self, notebook_path, args, inputs, outputs):
        self.__notebook_path = notebook_path
        self.__args = args
        self.__inputs = inputs
        self.__outputs = outputs

    @property
    def notebook_path(self):
        return self.__notebook_path

    @property
    def args(self):
        return self.__args

    @property
    def inputs(self):
        return self.__inputs

    @property
    def outputs(self):
        return self.__outputs


class PipelineState():
    """
    Represents the current execution state of notebooks in a pipeline.
    """

    def __init__(self):
        # Keeps last completed notebooks' runs
        self.__runs = {}

    @property
    def runs(self):
        """
        Returns a generator over added runs.
        """
        for run in self.__runs.values():
            yield run

    def record_run(self, run: NotebookRun):
        """
        Records a notebook's run.
        If a run has a notebook's path remembered before then it overwrites the latter.

        Parameters
        ----------
        run: NotebookRun
            A run of a notebook.
        """
        if run is None:
            raise ValueError("run cannot be None.")

        self.__runs[run.notebook_path] = run



from dtflw.storage.fs import FileStorageBase
from pyspark.sql.session import SparkSession
import dtflw.databricks as db


class AzureStorage(FileStorageBase):
    """
    Provides operations with files and directories on an Azure Storage container.
    """

    def __init__(self, account_name: str, container_name: str, root_dir: str, spark: SparkSession, dbutils):
        """
        Initializes an instance.

        Parameters
        ----------
        account_name : str
            Azure Storage account name.
        container_name : str
            Azure Storage container name.
        root_dir : str
            Root dir in a container.
        spark: SparkSession
            A Spark session object.
        dbutils: DBUtils
            A DBUtils object.
        """
        super().__init__(spark, dbutils, root_dir)

        self.__account_name = account_name
        self.__container_name = container_name

    @property
    def account_name(self) -> str:
        return self.__account_name

    @property
    def container_name(self) -> str:
        return self.__container_name

    @property
    def base_path(self):
        """
        Returns the base path.
        """
        return f"wasbs://{self.__container_name}@{self.__account_name}.blob.core.windows.net/"


def init_storage(account_name: str, container_name: str, root_dir: str = None, spark: SparkSession = None, dbutils=None) -> AzureStorage:
    """
    Returns a new instance of AzureStorage. 
    It is suggested using this factory function instead of the constructor of AzureStorage class.

    Parameters
    ----------
    account_name : str
        Azure Storage account name.
    container_name : str
        Azure Storage container name.
    root_dir : str (None)
        Root dir in a container.
        If None then a username of the current user will be used.
    spark: SparkSession (None)
        A Spark session object.
        If None then the current instance is used.
    dbutils: DBUtils (None)
        A DBUtils object.
        If None then the current instance is used.
    """
    if account_name is None or len(account_name) == 0:
        raise ValueError("account_name cannot be None nor empty string.")

    if container_name is None or len(container_name) == 0:
        raise ValueError("container_name cannot be None nor empty string.")

    if root_dir is None:
        root_dir = db.get_current_username()

    if spark is None:
        spark = db.get_spark_session()

    if dbutils is None:
        dbutils = db.get_dbutils()

    return AzureStorage(account_name, container_name, root_dir, spark, dbutils)



from dtflw.storage.fs import FileStorageBase
from pyspark.sql.session import SparkSession
import dtflw.databricks as db


class DbfsStorage(FileStorageBase):
    """
    Provides operations with files and directories on DBFS (the DBFS root).
    """
    def __init__(self, root_dir: str, spark: SparkSession, dbutils):

        super().__init__(spark, dbutils, root_dir)

    @property
    def base_path(self):
        return "dbfs:/"

def init_storage(root_dir: str = None, spark: SparkSession = None, dbutils=None) -> DbfsStorage:
    """
    Returns a new instance of DbfsStorage. 
    It is suggested using this factory function instead of the constructor of the class.

    Parameters
    ----------
    root_dir : str (None)
        Root dir in a container.
        If None then the 'dbfs:/FileStore/{username}' will be used.
    spark: SparkSession (None)
        A Spark session object.
        If None then the current instance is used.
    dbutils: DBUtils (None)
        A DBUtils object.
        If None then the current instance is used.
    """

    if root_dir is None:
        root_dir = f"FileStore/{db.get_current_username()}"

    if spark is None:
        spark = db.get_spark_session()

    if dbutils is None:
        dbutils = db.get_dbutils()

    return DbfsStorage(root_dir, spark, dbutils)



import unittest
from unittest.mock import MagicMock, patch
from dtflw.logger import DefaultLogger
from ddt import ddt, data, unpack
from dtflw.lazy_notebook import LazyNotebook
from dtflw.flow_context import FlowContext
from tests import utils
from tests.utils import StorageMock


@ddt
class LazyNotebookTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self._storage = StorageMock("", None, None)

        self._ctx = FlowContext(
            storage=self._storage,
            spark=None,
            dbutils=None,
            logger=DefaultLogger()
        )

    @data(
        # Input's abs path is None.
        (None, None),
        # Input's abs path gets resolved from earlier outputs.
        (None, "mock:/foo.txt"),
        # Input's abs path gets built from the given relative path.
        ("bar/foo.txt", "mock:/bar/foo.txt"),
        # Input's abs path is given explicitly.
        ("mock:/foo.txt", "mock:/foo.txt"),
        # Input's abs path gets resolved from the earlier output given by source_table.
        (None, "mock:/baz.txt", "baz")
    )
    @unpack
    def test_input(self, file_path, expected_input_path, source_table=None):

        # Arrange

        if file_path is None and expected_input_path is not None:
            # The input table is expected to be resolved as an output of an earlier notebook.
            if source_table is None:
                # Case 1: search for a table with input's name.
                self._ctx.publish_tables(
                    {"foo": expected_input_path}, "some_earlier_notebook")
            else:
                # Case 2: search for a table with provided name.
                self._ctx.publish_tables(
                    {source_table: expected_input_path}, "some_earlier_notebook")

        # Act
        nb = LazyNotebook("nb", self._ctx).input(
            name="foo",
            file_path=file_path,
            source_table=source_table
        )

        # Assert

        inputs = nb.get_inputs()

        self.assertEqual(len(inputs), 1)
        self.assertIn("foo", inputs)
        self.assertEqual(inputs["foo"].name, "foo")
        self.assertEqual(inputs["foo"].abs_file_path, expected_input_path)

    def test_input_name_none_fails(self):

        with self.assertRaises(ValueError):
            LazyNotebook("nb", self._ctx).input(name=None)

    @data(
        # Default. Output's abs file path gets resolved from a path of the current notebook.
        (None, "mock:/project/nb/foo.parquet"),
        # Outputs's path is given as a relative path.
        ("project/foo.parquet", "mock:/project/foo.parquet"),
        # Outputs's path is given as an abs file path.
        ("mock:/foo.parquet",
         "mock:/foo.parquet")
    )
    @unpack
    @patch("dtflw.databricks.get_this_notebook_abs_path")
    def test_output_abs_file_path(self, file_path, expected_output_path, get_this_notebook_abs_path_mock):

        # Arrange
        get_this_notebook_abs_path_mock.return_value = "/Repos/a@b.c/project/main"

        # Act
        nb = LazyNotebook("nb", self._ctx).output(
            name="foo",
            cols=None,
            file_path=file_path,
            alias=None
        )

        # Assert
        outputs = nb.get_outputs()

        self.assertEqual(len(outputs), 1)
        self.assertIn("foo", outputs)
        self.assertEqual(outputs["foo"].name, "foo")
        self.assertEqual(outputs["foo"].abs_file_path, expected_output_path)
        self.assertEqual(outputs["foo"].alias, "foo")

    @patch("dtflw.databricks.get_this_notebook_abs_path")
    def test_output_alias(self, get_this_notebook_abs_path_mock):

        # Arrange
        get_this_notebook_abs_path_mock.return_value = "/Repos/a@b.c/project/main"

        # Act
        nb = LazyNotebook("nb", self._ctx).output(
            name="foo",
            cols=None,
            file_path=None,
            alias="FooAliased"
        )

        # Assert
        outputs = nb.get_outputs()

        # Assert
        self.assertIn("foo", outputs)
        self.assertEqual(outputs["foo"].alias, "FooAliased")

    def test_output_name_none_fails(self):

        with self.assertRaises(ValueError):
            LazyNotebook("nb", self._ctx).output(name=None)

    @data(
        (True, False, False),
        (True, True, True),
        (False, False, True),
        (False, True, True),
    )
    @unpack
    @patch("dtflw.databricks.get_spark_session")
    @patch("dtflw.databricks.is_job_interactive")
    @patch("dtflw.databricks.get_this_notebook_abs_path")
    @patch("dtflw.output_table.OutputTable.needs_eval")
    @patch("dtflw.output_table.OutputTable.validate")
    @patch("dtflw.databricks.run_notebook")
    def test_run(self, is_lazy, outputs_need_eval, is_expected_running, run_notebook_mock: MagicMock, output_validate_mock, output_needs_eval_mock, get_this_notebook_abs_path_mock, is_job_interactive_mock, get_session_mock):
        """
            Flow does not run a notebook
                if is_lazy is True AND all outputs are evaluated.

            Flow runs a notebook
                if at least one output needs to be evaluated OR is_lazy is False.
        """
        # Arrange
        get_this_notebook_abs_path_mock.return_value = "/Repos/a@b.c/project/main"
        output_needs_eval_mock.return_value = outputs_need_eval
        is_job_interactive_mock.return_value = True
        get_session_mock.return_value = utils.SparkSessionMock()

        def do_nothing(strict):
            pass
        output_validate_mock.side_effect = do_nothing

        # Act

        nb = LazyNotebook("nb", self._ctx).output("foo")
        nb.run(is_lazy=is_lazy, strict_validation=False)

        if is_expected_running:
            run_notebook_mock.assert_called_once_with(
                nb.rel_path,
                0,
                {
                    "foo_out": self._ctx.storage.get_abs_path("project/nb/foo.parquet")
                }
            )
        else:
            run_notebook_mock.assert_not_called()

    @patch("dtflw.databricks.is_job_interactive")
    def test_run_failure_input_not_found(self, is_job_interactive_mock):
        """
        Raises an exeption if a required input table was not found
        before a run.
        """
        is_job_interactive_mock.return_value = False

        nb = LazyNotebook("nb", self._ctx).input("foo")

        with self.assertRaisesRegex(Exception, "Required input not found."):
            nb.run(is_lazy=False)

    @patch("dtflw.databricks.get_spark_session")
    @patch("dtflw.databricks.is_job_interactive")
    @patch("dtflw.databricks.get_this_notebook_abs_path")
    @patch("dtflw.output_table.OutputTable.needs_eval")
    @patch("dtflw.databricks.run_notebook")
    def test_run_failure_output_not_found(self, run_notebook_mock: MagicMock, output_needs_eval_mock, get_this_notebook_abs_path_mock, is_job_interactive_mock, get_session_mock):
        """
        Raises an exeption if an expected output table was not found
        after a run.
        """
        # Arrange
        get_this_notebook_abs_path_mock.return_value = "/Repos/a@b.c/project/main"
        output_needs_eval_mock.return_value = True
        is_job_interactive_mock.return_value = True
        get_session_mock.return_value = utils.SparkSessionMock()

        def do_nothing(path: str, timeout: int, args: dict):
            pass

        run_notebook_mock.side_effect = do_nothing

        nb = LazyNotebook("nb", self._ctx).output("foo")

        # Act/Assert
        with self.assertRaisesRegex(Exception, "Expected output not found."):
            nb.run(is_lazy=False)

    @patch("dtflw.databricks.get_spark_session")
    @patch("dtflw.databricks.is_job_interactive")
    @patch("dtflw.databricks.get_this_notebook_abs_path")
    @patch("dtflw.databricks.run_notebook")
    def test_run_return_value(self, run_notebook_mock: MagicMock, get_this_notebook_abs_path_mock, is_job_interactive_mock, get_session_mock):
        """
        Raises an exeption if an expected output table was not found
        after a run.
        """

        # Arrange
        get_this_notebook_abs_path_mock.return_value = "/Repos/a@b.c/project/main"
        run_notebook_mock.return_value = "foo"
        is_job_interactive_mock.return_value = True
        get_session_mock.return_value = utils.SparkSessionMock()

        actual = LazyNotebook("nb", self._ctx).run(is_lazy=False)

        # Act/Assert
        self.assertEqual(actual, "foo")



from abc import ABC, abstractmethod


class FlowPluginBase(ABC):
    """
    Inherit from this class to implement a Flow plugin.
    """
    @property
    @abstractmethod
    def action_name(self) -> str:
        """
        Name of a method that will be attached to a Flow instance.
        """
        raise NotImplementedError()

    @abstractmethod
    def act(self, flow, *args, **kwargs):
        """
        Implementation of the method.

        Parameters
        ----------
        flow : Flow
            Flow instance to which this plugin is installed.
        """
        raise NotImplementedError()


class NotebookPluginBase(ABC):
    """
    Inherit from this class to implement a LazyNotebook plugin.
    """
    @property
    @abstractmethod
    def action_name(self) -> str:
        """
        Name of a method that will be attached to LazyNotebook instances.
        """
        raise NotImplementedError()

    @abstractmethod
    def act(self, notebook, flow, *args, **kwargs):
        """
        Implementation of the method.

        Parameters
        ----------
        flow : Flow
            Flow instance which created the current notebook.

        notebook : LazyNotebook
            Instance of LazyNotebook to which this plugin is installed. 

        """
        raise NotImplementedError()



import unittest
from dtflw.logger import DefaultLogger
import sys
from io import StringIO
from ddt import ddt, data, unpack


@ddt
class DefaultLoggerTestCase(unittest.TestCase):

    def test_verbosity_set_unknown_value(self):

        logger = DefaultLogger()
        logger.verbosity = "unknown value"

        self.assertEqual(logger.verbosity, "default")

    @data(
        ("verbose", "foo", "foo"),
        ("default", "foo", "")
    )
    @unpack
    def test_log(self, verbosity, info_msg, expected_msg):

        logger = DefaultLogger(verbosity)

        out = StringIO()
        sys.stdout = out
        logger.log(info_msg)

        actual_msg = out.getvalue().strip()

        # Assert
        self.assertEqual(actual_msg, expected_msg)



from collections import OrderedDict
import typing


class TablesRepository:
    """
    Repository to keep relations between notebooks and tables they produce.
    """

    def __init__(self):
        self.__tables = {}

    @property
    def tables(self):
        """
        Published tables.

        Returns
        -------
        A dictionary of structure below
        { 
            <table name> : { 
                <notebook path> : <table abs path on a storage>
                } 
        }
        """
        return self.__tables

    def publish(self, tables: typing.Dict[str, str], src_nb_path: str) -> None:
        """
        Registers tables as being available to be resolved for a notebook as input.

        Parameters
        ----------
        tables: Dict[str, str] 
            A dictionary where key is a table's name and value is an absolute file path.

        src_nb_path: str
            Path of a notebook which has produced tables (a source notebook).
        """
        for t_name, t_abs_path in tables.items():
            if t_name not in self.__tables:
                self.__tables[t_name] = OrderedDict()

            self.__tables[t_name][src_nb_path] = t_abs_path

    def resolve(self, name: str, trg_nb_path: str = None) -> str:
        """
        Searches for a table to be returned as an input for a notebook.
        Returns None if no suitable table was found.

        Table Search Algorithm
        1. IF the target notebook has not published tables yet
            THEN return the latest table path available.
        2. IF the target notebook has already published tables
            THEN return the first table path evalueted right before it.
        3. IF the target notebook is None
            THEN return the latest table published by a given name 
            (if exsts otherwise return None).

        Parameters
        ----------
        name: str
            Name of a table to search for.
        trg_nb_path: str
            Path of a notebook for which a table will be resolved as an input (a target notebook).

        Returns
        -------
        str
            Absolute path to the table if it has been found, and None otherwise.
        """

        if name not in self.__tables:
            return None

        if not trg_nb_path or trg_nb_path not in self.__tables[name]:
            return next(reversed(self.__tables[name].values()))

        # Iterate backwards to pick the latest table published.
        is_found = False
        for t_nb_path, t_abs_path in reversed(self.__tables[name].items()):
            if is_found:
                return t_abs_path
            is_found = trg_nb_path == t_nb_path

        return None



from __future__ import annotations
from dtflw.pipeline import NotebookRun
import dtflw.databricks as db
import typing
from dtflw.flow_context import FlowContext
from dtflw.input_table import InputTable
from dtflw.output_table import OutputTable
import dtflw.arguments as a
import dtflw.com as com


class LazyNotebook:

    class AliasedOutputTable(OutputTable):
        """
        Output table with an alias.
        """

        def __init__(self, name: str, abs_file_path: str, cols: list, ctx: FlowContext, alias: str):
            super().__init__(name, abs_file_path, cols, ctx)
            self.alias = alias

    def __init__(self, rel_path: str, ctx: FlowContext):
        self.__rel_path = rel_path
        self.__args = {}
        self.__timeout = 0

        self.ctx = ctx

        self.__inputs = {}
        self.__outputs = {}

        self.__last_run_result = None

    @property
    def rel_path(self):
        return self.__rel_path

    def args(self, args: typing.Dict[str, str]) -> LazyNotebook:
        self.__args = args
        return self

    def get_args(self):
        """
        Returns arguments.
        """
        return self.__args

    def timeout(self, timeout: int) -> LazyNotebook:
        self.__timeout = timeout
        return self

    def input(self, name: str, file_path: str = None, source_table: str = None) -> LazyNotebook:
        """
        Registers an input table which is required by the notebook to run.

        Parameters
        ----------
        name: str
            Name of an input table.
        file_path: str = None
            File path to an input table.
            - If None (default) then input path is gets resolved.
            - If given as a relative path then gets changed to absolute.
            - If given as an absolute path then it is used as given.
        source_table: str
            Name of the source table to pass an an input. 
            If 'source_table' is not given then 'name' is used.
        """
        if name is None or len(name) == 0:
            raise ValueError("Input's name cannot be empty.")

        input_file_path = file_path
        if input_file_path is None:
            # Resolve the input's file either by given 'source' or its `name`
            source_table_name = name
            if source_table is not None and len(source_table) > 0:
                source_table_name = source_table

            input_file_path = self.ctx.resolve_table(
                source_table_name,
                self.rel_path
            )

        elif not self.ctx.storage.is_abs_path(input_file_path):
            # Bind the input to a given specific file
            input_file_path = self.ctx.storage.get_abs_path(input_file_path)

        self.__inputs[name] = InputTable(
            name,
            input_file_path,
            self.ctx)
        return self

    def get_inputs(self):
        """
        Returns input tables.
        """
        return self.__inputs

    def output(self, name: str, cols: list = None, file_path: str = None, alias: str = None) -> LazyNotebook:
        """
        Registers an output table which is expected to be produced by the notebook.

        Parameters
        ----------
        name: str
            Name of an output table.
        cols: list[(str,str)]
            Schema of an expected output (DataFrame.dtypes format).
        file_path: str = None
            File path to an output table.
            - If None (default) then output path is built from notebook's path and output's name.
            - If given as a relative path then gets changed to absolute.
            - If given as an absolute path then it is used as given.
        alias: str
            Alias by which the output is registered in Flow. If not specified then 'name' value is used.
        """
        if name is None or len(name) == 0:
            raise ValueError("Output's name cannot be empty.")

        output_file_path = file_path

        if output_file_path is None:
            # default behavior

            # path of this notebook starting with project's dir
            project_based_nb_path = db.get_path_relative_to_project_dir(
                self.rel_path)

            output_file_path = self.ctx.storage.get_abs_path(
                self.ctx.storage.get_path_in_root_dir(
                    self.ctx.storage.get_path_with_file_extension(
                        f"{project_based_nb_path}/{name}"
                    )
                )
            )
        elif not self.ctx.storage.is_abs_path(output_file_path):
            output_file_path = self.ctx.storage.get_abs_path(output_file_path)

        self.__outputs[name] = self.AliasedOutputTable(
            name,
            output_file_path,
            cols,
            self.ctx,
            alias if alias is not None and len(alias) > 0 else name)

        return self

    def get_outputs(self):
        """
        Returns output tables.
        """
        return self.__outputs

    def __validate_tables(self, tables, title: str, strict: bool = False) -> typing.Dict[str, str]:
        """
        Validates if an input/output table is valid.
        """
        for t in tables:
            self.ctx.logger.log(f"{title} '{t.name}': ")

            t.validate(strict)
            self.ctx.logger.log(f"\t'{t.abs_file_path}'")

    def share_arguments(self) -> None:
        """
        Makes values of args, inputs and outputs available in the callee notebook.
        Retrieve them with `init_args`, `init_inputs` and `init_outputs` respectively.
        """

        ch = com.NotebooksChannel()

        ch.share_args(
            db.get_notebook_abs_path(self.rel_path),
            self.get_args()
        )

        ch.share_inputs(
            db.get_notebook_abs_path(self.rel_path),
            {name: i.abs_file_path for name, i in self.get_inputs().items()}
        )

        ch.share_outputs(
            db.get_notebook_abs_path(self.rel_path),
            {name: o.abs_file_path for name, o in self.get_outputs().items()}
        )

    def run(self, is_lazy: bool = False, strict_validation: bool = False):
        """
        Runs the notebook.

        Parameters
        ----------
        is_lazy: bool
            Lazy behavior depending on the condition
            is_lazy AND (at least one output needs to be evaluated)

            True, False  => skip
            True, True   => run
            False, False => run
            False, True  => run

        strict_validation: bool
            If True then raises an exception if output tables are not exactly as expected.
            Otherwise, allows an output to have more columns than expected.

        Returns
        -------
        Argument of the `dbutils.notebook.exit` if exists otherwise `None`.
        """

        if db.is_job_interactive():
            # If interactive then share arguments for easy debugging of the callee notebook.
            self.share_arguments()

        # Check if inputs are valid. Raises an exception if not.
        self.__validate_tables(self.__inputs.values(), "Input")

        any_output_needs_eval = any(
            [o.needs_eval() for o in self.__outputs.values()]
        )

        if is_lazy and not any_output_needs_eval:
            self.ctx.logger.log(
                f"Skipped run: '{db.get_notebook_abs_path(self.rel_path)}'")

        else:
            self.ctx.logger.log(
                f"Running: '{db.get_notebook_abs_path(self.rel_path)}'")

            arguments = {
                **{a.Argument.get_full_name(name): value for name, value in self.get_args().items()},
                **{a.Input.get_full_name(name): i.abs_file_path for name, i in self.get_inputs().items()},
                **{a.Output.get_full_name(name): o.abs_file_path for name, o in self.get_outputs().items()}
            }

            self.__last_run_result = db.run_notebook(
                self.rel_path,
                self.__timeout,
                arguments
            )

        # Check if outputs are valid. Raises an exception if not.
        self.__validate_tables(
            self.__outputs.values(),
            "Output",
            strict=strict_validation
        )

        outputs = {o.alias: o.abs_file_path for o in self.__outputs.values()}
        self.ctx.publish_tables(outputs, self.rel_path)

        # Record the completed notebook's run.
        inputs = {name: i.abs_file_path for name, i in self.__inputs.items()}

        self.ctx.pipeline.record_run(
            NotebookRun(
                db.get_notebook_abs_path(self.rel_path),
                self.get_args(),
                inputs,
                outputs
            )
        )

        return self.__last_run_result

    def show(self):
        """
        Prints the current state of the notebook.
        """

        c = f"Notebook: '{self.rel_path}'"
        c += "\nArgs:"
        for (a, v) in self.__args.items():
            c += f"\n\t'{a}': '{v}'"

        c += "\nInputs:"
        for (a, i) in self.__inputs.items():
            c += f"\n\t'{a}': '{i.abs_file_path}'"

        c += "\nOutputs:"
        for (a, o) in self.__outputs.items():
            c += f"\n\t'{a}': '{o.abs_file_path}'"

        self.ctx.display.show(c)
